[
  # Initially all files are ignored
  # TODO: When fixing / completing specs & types of files, please remove the
  # corresponding entry below
  {"lib/jose/jwa.ex", :unknown_type},
  {"lib/jose/jwa.ex", :invalid_contract},
  {"lib/jose/jwe.ex", :call},
  {"lib/jose/jwe.ex", :pattern_match_cov},
  {"lib/jose/jwe.ex", :invalid_contract},
  {"lib/jose/jwk.ex", :call},
  {"lib/jose/jwk.ex", :pattern_match_cov},
  {"lib/jose/jws.ex", :pattern_match_cov},
  {"lib/jose/jwt.ex", :call},
  {"lib/jose/jwt.ex", :guard_fail},
  {"src/jose_public_key.erl", :pattern_match},
  {"src/jose_public_key.erl", :pattern_match_cov},
  {"src/jose_public_key.erl", :call},
  {"src/jose_public_key.erl", :no_return},
  {"src/jose_server.erl", :pattern_match},
  {"src/json/jose_json_jiffy.erl", :unknown_function},
  {"src/jwa/curve25519/jose_curve25519_crypto.erl", :callback_arg_type_mismatch},
  {"src/jwa/curve448/jose_curve448_crypto.erl", :callback_arg_type_mismatch},
  {"src/jwa/curve448/jose_curve448_fallback.erl", :callback_spec_arg_type_mismatch},
  {"src/jwa/curve448/jose_curve448_fallback.erl", :callback_spec_type_mismatch},
  {"src/jwa/jose_jwa_curve25519.erl", :callback_arg_type_mismatch},
  {"src/jwa/jose_jwa_curve448.erl", :callback_arg_type_mismatch},
  {"src/jwa/jose_jwa_curve448.erl", :callback_type_mismatch},
  {"src/jwa/jose_jwa_pkcs1.erl", :invalid_contract},
  {"src/jwa/jose_jwa_pkcs1.erl", :pattern_match_cov},
  {"src/jwa/jose_jwa_pkcs1.erl", :guard_fail},
  {"src/jwa/sha3/jose_sha3_keccakf1600_driver.erl", :unknown_function},
  {"src/jwa/sha3/jose_sha3_keccakf1600_nif.erl", :unknown_function},
  {"src/jwe/jose_jwe_alg_ecdh_1pu.erl", :record_match},
  {"src/jwe/jose_jwe_alg_ecdh_1pu.erl", :record_constr},
  {"src/jwe/jose_jwe_alg_ecdh_es.erl", :record_match},
  {"src/jwe/jose_jwe_alg_ecdh_es.erl", :record_constr},
  {"src/jwe/jose_jwe_alg_ecdh_ss.erl", :record_match},
  {"src/jwe/jose_jwe_alg_ecdh_ss.erl", :record_constr},
  {"src/jwk/jose_jwk_kty_okp_ed25519.erl", :call},
  {"src/jwk/jose_jwk_kty_okp_ed25519ph.erl", :call},
  {"src/jwk/jose_jwk_kty_okp_ed448.erl", :call},
  {"src/jwk/jose_jwk_kty_okp_ed448ph.erl", :call},
  {"src/jwk/jose_jwk_kty_rsa.erl", :unknown_function},
  {"src/jws/jose_jws_alg_poly1305.erl", :call},
  {"src/jws/jose_jws_alg_poly1305.erl", :record_constr},
  {"src/jws/jose_jws_alg_poly1305.erl", :guard_fail_pat},
  {"src/jws/jose_jws_alg_poly1305.erl", :no_return}
]
